/* -*- Mode: C++; tab-width: 2; indent-tabs-mode: nil; c-basic-offset: 4 -*- */
/* ***** BEGIN LICENSE BLOCK *****
 * Version: MPL 1.1/GPL 2.0/LGPL 2.1
 *
 * The contents of this file are subject to the Mozilla Public License Version
 * 1.1 (the "License"); you may not use this file except in compliance with
 * the License. You may obtain a copy of the License at
 * http://www.mozilla.org/MPL/
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The Original Code is mozilla.org code.
 *
 * The Initial Developer of the Original Code is
 * Netscape Communications Corporation.
 * Portions created by the Initial Developer are Copyright (C) 2001
 * the Initial Developer. All Rights Reserved.
 *
 * Contributor(s):
 *   Seth Spitzer <sspitzer@netscape.com>
 *   Postbox, Inc. <contact@postbox-inc.com>
 *
 * Alternatively, the contents of this file may be used under the terms of
 * either of the GNU General Public License Version 2 or later (the "GPL"),
 * or the GNU Lesser General Public License Version 2.1 or later (the "LGPL"),
 * in which case the provisions of the GPL or the LGPL are applicable instead
 * of those above. If you wish to allow use of your version of this file only
 * under the terms of either the GPL or the LGPL, and not to allow others to
 * use your version of this file under the terms of the MPL, indicate your
 * decision by deleting the provisions above and replace them with the notice
 * and other provisions required by the GPL or the LGPL. If you do not delete
 * the provisions above, a recipient may use your version of this file under
 * the terms of any one of the MPL, the GPL or the LGPL.
 *
 * ***** END LICENSE BLOCK ***** */

#include "msgCore.h"
#include "nsMsgSearchDBView.h"
#include "nsIMsgHdr.h"
#include "nsIMsgThread.h"
#include "nsQuickSort.h"
#include "nsIDBFolderInfo.h"
#include "nsMsgBaseCID.h"
#include "nsIMsgCopyService.h"
#include "nsICopyMsgStreamListener.h"
#include "nsMsgUtils.h"
#include "nsITreeColumns.h"
#include "nsIMsgMessageService.h"
#include "nsMsgGroupThread.h"
#include "nsIPrefService.h"
#include "nsIPrefBranch.h"

static PRBool gReferenceOnlyThreading;
#ifndef NO_POSTBOX
static PRBool gThreadWithoutRe;
#endif

nsMsgSearchDBView::nsMsgSearchDBView()
{
  // don't try to display messages for the search pane.
  mSuppressMsgDisplay = PR_TRUE;
  m_threadsTable.Init();
  m_hdrsTable.Init();
}

nsMsgSearchDBView::~nsMsgSearchDBView()
{	
}

NS_IMPL_ISUPPORTS_INHERITED2(nsMsgSearchDBView, pbConversationView, nsIMsgDBView, nsIMsgSearchNotify)

NS_IMETHODIMP nsMsgSearchDBView::Open(nsIMsgFolder *folder, 
                                      nsMsgViewSortTypeValue sortType, 
                                      nsMsgViewSortOrderValue sortOrder, 
                                      nsMsgViewFlagsTypeValue viewFlags, 
                                      PRInt32 *pCount)
{
  // dbViewWrapper.js likes to create search views with a sort order
  // of byNone, in order to have the order be the order the search results
  // are returned. But this doesn't work with threaded view, so make the
  // sort order be byDate if we're threaded.

  if (viewFlags & nsMsgViewFlagsType::kThreadedDisplay &&
      sortType == nsMsgViewSortType::byNone)
    sortType = nsMsgViewSortType::byDate;

  nsresult rv = nsMsgDBView::Open(folder, sortType, sortOrder, 
                                    viewFlags, pCount);
  NS_ENSURE_SUCCESS(rv, rv);

  nsCOMPtr<nsIPrefBranch> prefBranch(do_GetService(NS_PREFSERVICE_CONTRACTID, &rv));
  NS_ENSURE_SUCCESS(rv, rv);
  prefBranch->GetBoolPref("mail.strict_threading", &gReferenceOnlyThreading);
#ifndef NO_POSTBOX
  prefBranch->GetBoolPref("mail.thread_without_re", &gThreadWithoutRe);
#endif
  // our sort is automatically valid because we have no contents at this point!
  m_sortValid = PR_TRUE;

    if (pCount)
      *pCount = 0;
    m_folder = nsnull;
    return rv;
}

PLDHashOperator
nsMsgSearchDBView::ThreadTableCloner(const nsAString &aKey, nsIMsgThread* aThread, void* aArg)
{
  nsMsgSearchDBView* view = static_cast<nsMsgSearchDBView*>(aArg);
  nsresult rv = view->m_threadsTable.Put(aKey, aThread);
  return NS_SUCCEEDED(rv) ? PL_DHASH_NEXT : PL_DHASH_STOP;
}

PLDHashOperator
nsMsgSearchDBView::MsgHdrTableCloner(const nsAString &aKey, nsIMsgDBHdr* aMsgHdr, void* aArg)
{
  nsMsgSearchDBView* view = static_cast<nsMsgSearchDBView*>(aArg);
  nsresult rv = view->m_hdrsTable.Put(aKey, aMsgHdr);
  return NS_SUCCEEDED(rv) ? PL_DHASH_NEXT : PL_DHASH_STOP;
}

NS_IMETHODIMP
nsMsgSearchDBView::CopyDBView(nsMsgDBView *aNewMsgDBView, nsIMessenger *aMessengerInstance, 
                                       nsIMsgWindow *aMsgWindow, nsIMsgDBViewCommandUpdater *aCmdUpdater)
{
  nsMsgGroupView::CopyDBView(aNewMsgDBView, aMessengerInstance, aMsgWindow, aCmdUpdater);
  nsMsgSearchDBView* newMsgDBView = (nsMsgSearchDBView *) aNewMsgDBView;

  // now copy all of our private member data
  newMsgDBView->mDestFolder = mDestFolder;
  newMsgDBView->mCommand = mCommand;
  newMsgDBView->mTotalIndices = mTotalIndices;
  newMsgDBView->mCurIndex = mCurIndex; 
  newMsgDBView->m_folders.InsertObjectsAt(m_folders, 0);

  if (m_hdrsForEachFolder)
    m_hdrsForEachFolder->Clone(getter_AddRefs(newMsgDBView->m_hdrsForEachFolder));

  if (m_copyListenerList)
    m_copyListenerList->Clone(getter_AddRefs(newMsgDBView->m_copyListenerList));

  newMsgDBView->m_uniqueFoldersSelected.InsertObjectsAt(m_uniqueFoldersSelected, 0);

  PRInt32 count = m_dbToUseList.Count(); 
  for(PRInt32 i = 0; i < count; i++)
  {
    newMsgDBView->m_dbToUseList.AppendObject(m_dbToUseList[i]);
    // register the new view with the database so it gets notifications
    m_dbToUseList[i]->AddListener(newMsgDBView);
  }

  if (m_viewFlags & nsMsgViewFlagsType::kThreadedDisplay)
  {
    // We need to clone the thread and msg hdr hash tables.
    m_threadsTable.EnumerateRead(ThreadTableCloner, newMsgDBView);
    m_hdrsTable.EnumerateRead(MsgHdrTableCloner, newMsgDBView);
  }

  return NS_OK;
}

NS_IMETHODIMP nsMsgSearchDBView::Close()
{
  PRInt32 count = m_dbToUseList.Count();
  
  for(PRInt32 i = 0; i < count; i++)
    m_dbToUseList[i]->RemoveListener(this);

  m_dbToUseList.Clear();

  return nsMsgGroupView::Close();
}

void nsMsgSearchDBView::InternalClose()
{
  m_threadsTable.Clear();
  m_hdrsTable.Clear();
  nsMsgGroupView::InternalClose();
  m_folders.Clear();
}

NS_IMETHODIMP nsMsgSearchDBView::GetCellText(PRInt32 aRow, nsITreeColumn* aCol, nsAString& aValue)
{
  if (!IsValidIndex(aRow))
    return NS_MSG_INVALID_DBVIEW_INDEX;

  const PRUnichar* colID;
  aCol->GetIdConst(&colID);
  // location, need to check for "lo" not just "l" to avoid "label" column
  if (colID[0] == 'l' && colID[1] == 'o') 
    return FetchLocation(aRow, aValue);
  else
    return pbConversationView::GetCellText(aRow, aCol, aValue);
}

nsresult nsMsgSearchDBView::FetchLocation(PRInt32 aRow, nsAString& aLocationString)
{
  nsCOMPtr <nsIMsgFolder> folder;
  nsresult rv = GetFolderForViewIndex(aRow, getter_AddRefs(folder));
  NS_ENSURE_SUCCESS(rv,rv);
  return folder->GetPrettiestName(aLocationString);
}

nsresult nsMsgSearchDBView::OnNewHeader(nsIMsgDBHdr *newHdr, nsMsgKey aParentKey,
                                        PRBool /*ensureListed*/)
{
   return NS_OK;
}

NS_IMETHODIMP 
nsMsgSearchDBView::OnHdrDeleted(nsIMsgDBHdr *aHdrDeleted, nsMsgKey aParentKey, 
                                PRInt32 aFlags, nsIDBChangeListener *aInstigator)
{
  if (m_viewFlags & nsMsgViewFlagsType::kGroupBySort)
    return nsMsgGroupView::OnHdrDeleted(aHdrDeleted, aParentKey, 
                                        aFlags, aInstigator);

  DeleteEditingTaskMsgHdr(aHdrDeleted);

  if (m_viewFlags & nsMsgViewFlagsType::kThreadedDisplay)
  {
#ifndef NO_POSTBOX
    nsMsgViewIndex deletedIndex = FindHdr(aHdrDeleted);
    if (deletedIndex != nsMsgViewIndex_None)
    {
      // if the message deleted is a dummy header, find the real header.
      if ((m_flags[deletedIndex] & MSG_VIEW_FLAG_DUMMY) && ! (m_flags[deletedIndex] & MSG_FLAG_ELIDED))
      {
        for (deletedIndex = deletedIndex + 1; deletedIndex < (PRUint32) GetSize(); deletedIndex++)
        {
          nsCOMPtr<nsIMsgDBHdr> tempHdr;
          GetMsgHdrForViewIndex(deletedIndex, getter_AddRefs(tempHdr));
          if (tempHdr == aHdrDeleted)
            break;
        }
      }
      nsCOMPtr<nsIMsgDBHdr> tmpHdr;
      GetMsgHdrForViewIndex(deletedIndex, getter_AddRefs(tmpHdr));
      if (tmpHdr == aHdrDeleted)
      {
        // don't directly call nsMsgSearchDBView::RemoveByIndex as that sets the level of the remaining
        // message to 0 if there are only two messages in the conversation.
        m_folders.RemoveObjectAt(deletedIndex);
        nsMsgDBView::RemoveByIndex(deletedIndex);
        if (m_deletingRows)
          mIndicesToNoteChange.Add(deletedIndex);
      }
    }

    nsCOMPtr<nsIMsgThread> thread;
    nsCOMPtr<nsIMsgDBHdr> rootHdr;
    GetXFThreadFromMsgHdr(aHdrDeleted, getter_AddRefs(thread));
    if (thread)
    {
      nsMsgViewIndex viewIndexOfThread = GetIndexOfFirstDisplayedKeyInThread(thread);
      GetMsgHdrForViewIndex(viewIndexOfThread, getter_AddRefs(rootHdr));

      nsMsgXFViewThread *viewThread = static_cast<nsMsgXFViewThread*>(thread.get());
      viewThread->RemoveChildHdr(aHdrDeleted, nsnull);
      
      PRUint32 numChildren;
      thread->GetNumChildren(&numChildren);
      if (numChildren < 2)
      {
        if (viewIndexOfThread != nsMsgViewIndex_None)
        {
          if (deletedIndex != nsMsgViewIndex_None)
          {
            RemoveByIndex(viewIndexOfThread);
            if (m_deletingRows)
              mIndicesToNoteChange.Add(viewIndexOfThread);
            m_levels[viewIndexOfThread] = 0;
          }

          AndExtraFlag(viewIndexOfThread, ~(MSG_VIEW_FLAG_ISTHREAD | MSG_VIEW_FLAG_DUMMY | MSG_VIEW_FLAG_HASCHILDREN));
          OrExtraFlag(viewIndexOfThread, MSG_FLAG_ELIDED); // a thread with 1 child must be closed

          // get the hdr for the last remaining message in the thread
          GetMsgHdrForViewIndex(viewIndexOfThread, getter_AddRefs(rootHdr));
          if (rootHdr)
          {
            nsMsgKey rootKey;
            rootHdr->GetMessageKey(&rootKey);
            PRUint32 msgFlags;
            rootHdr->GetFlags(&msgFlags);
            
            // update the keys for the rich rows
            if (viewIndexOfThread && (m_viewFlags & nsMsgViewFlagsType::kRichView) && IsRichRow(viewIndexOfThread - 1))
            {
              // FIX ME if we add support for more than one rich row
              PRInt32 richRowFlags = msgFlags | MSG_VIEW_FLAG_TITLE_ROW;
              richRowFlags  &= ~(MSG_FLAG_ELIDED | MSG_VIEW_FLAG_HASCHILDREN | MSG_VIEW_FLAG_ISTHREAD | MSG_VIEW_FLAG_DUMMY);
              SetMsgHdrAt(rootHdr, viewIndexOfThread - 1, rootKey, richRowFlags, 0);
            }
          }
        }
        else 
        {
          // We just deleted the last message in the thread so remove the container row. We need to rely on deletedIndex instead of
          // viewIndexOfThread because viewIndexOfThread is nsMsgViewIndex_None
          // This needs fixed if we ever support more than one rich row per message
          if (deletedIndex && deletedIndex != nsMsgViewIndex_None && (m_viewFlags & nsMsgViewFlagsType::kRichView) && 
              IsRichRow(deletedIndex - 1))
          {
            m_folders.RemoveObjectAt(deletedIndex - 1);
            nsMsgDBView::RemoveByIndex(deletedIndex - 1);
            if (m_deletingRows)
              mIndicesToNoteChange.Add(deletedIndex - 1);
          }
        }

      }
      else if (rootHdr == aHdrDeleted && viewIndexOfThread != nsMsgKey_None)
      {
        // Promote the next best message to the top
        nsresult rv = GetThreadRootHdr(thread, getter_AddRefs(rootHdr));
        NS_ENSURE_SUCCESS(rv, rv);
        if (rootHdr)
        {
          nsMsgKey rootKey;
          rootHdr->GetMessageKey(&rootKey);
          PRUint32 msgFlags;
          rootHdr->GetFlags(&msgFlags);
          
          // update the keys for the rich rows
          if (viewIndexOfThread && (m_viewFlags & nsMsgViewFlagsType::kRichView) && IsRichRow(viewIndexOfThread - 1))
          {
            // FIX ME if we add support for more than one rich row
            PRInt32 richRowFlags = msgFlags | MSG_VIEW_FLAG_TITLE_ROW;
            richRowFlags  &= ~(MSG_FLAG_ELIDED | MSG_VIEW_FLAG_HASCHILDREN | MSG_VIEW_FLAG_ISTHREAD | MSG_VIEW_FLAG_DUMMY);
            SetMsgHdrAt(rootHdr, viewIndexOfThread - 1, rootKey, richRowFlags, 0);
          }
          SetMsgHdrAt(rootHdr, viewIndexOfThread, 
                              rootKey, msgFlags, 0);
          OrExtraFlag(viewIndexOfThread, MSG_VIEW_FLAG_ISTHREAD | MSG_VIEW_FLAG_DUMMY | MSG_VIEW_FLAG_HASCHILDREN);
        }
      }
    }
#else
    nsMsgViewIndex deletedIndex = FindHdr(aHdrDeleted);
    PRUint32 savedFlags = 0;
    if (deletedIndex != nsMsgViewIndex_None)
    {
      savedFlags = m_flags[deletedIndex];
      RemoveByIndex(deletedIndex);
    }

    nsCOMPtr<nsIMsgThread> thread;
    GetXFThreadFromMsgHdr(aHdrDeleted, getter_AddRefs(thread));
    if (thread)
    {
      nsMsgXFViewThread *viewThread = static_cast<nsMsgXFViewThread*>(thread.get());
      viewThread->RemoveChildHdr(aHdrDeleted, nsnull);
      if (deletedIndex == nsMsgViewIndex_None && viewThread->MsgCount() == 1)
      {
        // remove the last child of a collapsed thread. Need to find the root,
        // and remove the thread flags on it.
        nsCOMPtr<nsIMsgDBHdr> rootHdr;
        thread->GetRootHdr(nsnull, getter_AddRefs(rootHdr));
        if (rootHdr)
        {
          nsMsgViewIndex threadIndex = GetThreadRootIndex(rootHdr);
          if (threadIndex != nsMsgViewIndex_None)
            AndExtraFlag(threadIndex, ~(MSG_VIEW_FLAG_ISTHREAD | 
                                        MSG_FLAG_ELIDED | 
                                        MSG_VIEW_FLAG_HASCHILDREN));
        }
      }
      else if (savedFlags & MSG_VIEW_FLAG_HASCHILDREN)
      {
        if (savedFlags & MSG_FLAG_ELIDED)
        {
          nsCOMPtr<nsIMsgDBHdr> rootHdr;
          nsresult rv = thread->GetRootHdr(nsnull, getter_AddRefs(rootHdr));
          NS_ENSURE_SUCCESS(rv, rv);
          nsMsgKey msgKey;
          PRUint32 msgFlags;
          rootHdr->GetMessageKey(&msgKey);
          rootHdr->GetFlags(&msgFlags);
          // promote the new thread root
          if (viewThread->MsgCount() > 1)
            msgFlags |= MSG_VIEW_FLAG_ISTHREAD | MSG_FLAG_ELIDED | 
                        MSG_VIEW_FLAG_HASCHILDREN;
          InsertMsgHdrAt(deletedIndex, rootHdr, msgKey, msgFlags, 0);
          if (!m_deletingRows)
            NoteChange(deletedIndex, 1, nsMsgViewNotificationCode::insertOrDelete);
        }
        else if (viewThread->MsgCount() > 1)
        {
          OrExtraFlag(deletedIndex, MSG_VIEW_FLAG_ISTHREAD |
                                    MSG_VIEW_FLAG_HASCHILDREN);
        }
      }
    }
#endif
  }
  else
  {
    return nsMsgDBView::OnHdrDeleted(aHdrDeleted, aParentKey, 
                                        aFlags, aInstigator);
  }
   return NS_OK;
}
                                       
NS_IMETHODIMP nsMsgSearchDBView::OnHdrFlagsChanged(nsIMsgDBHdr *aHdrChanged, PRUint32 aOldFlags,
                                      PRUint32 aNewFlags, nsIDBChangeListener *aInstigator)
{
  // defer to base class if we're grouped or not threaded at all
  if (m_viewFlags & nsMsgViewFlagsType::kGroupBySort ||
      !(m_viewFlags && nsMsgViewFlagsType::kThreadedDisplay))
    return pbConversationView::OnHdrFlagsChanged(aHdrChanged, aOldFlags, 
                                             aNewFlags, aInstigator);

  nsCOMPtr <nsIMsgThread> thread;
  PRBool foundMessageId;
  // check if the hdr that changed is in a xf thread, and if the read flag
  // changed, update the thread unread count. GetXFThreadFromMsgHdr returns
  // the thread the header does or would belong to, so we need to also
  // check that the header is actually in the thread.
  GetXFThreadFromMsgHdr(aHdrChanged, getter_AddRefs(thread), &foundMessageId);
  if (foundMessageId)
  {
    nsMsgXFViewThread *viewThread = static_cast<nsMsgXFViewThread*>(thread.get());
    if (viewThread->HdrIndex(aHdrChanged) != -1)
    {
      PRUint32 deltaFlags = (aOldFlags ^ aNewFlags);
      if (deltaFlags & MSG_FLAG_READ)
        thread->MarkChildRead(aNewFlags & MSG_FLAG_READ);
    }
  }
  return nsMsgDBView::OnHdrFlagsChanged(aHdrChanged, aOldFlags,
                                        aNewFlags, aInstigator);
}

void nsMsgSearchDBView::InsertMsgHdrAt(nsMsgViewIndex index, nsIMsgDBHdr *hdr,
                              nsMsgKey msgKey, PRUint32 flags, PRUint32 level)
{
  m_keys.InsertAt(index, msgKey);
  m_flags.InsertAt(index, flags);
  m_levels.InsertAt(index, level);
  nsCOMPtr<nsIMsgFolder> folder;
  hdr->GetFolder(getter_AddRefs(folder));
  m_folders.InsertObjectAt(folder, index);
}

void nsMsgSearchDBView::SetMsgHdrAt(nsIMsgDBHdr *hdr, nsMsgViewIndex index, 
                              nsMsgKey msgKey, PRUint32 flags, PRUint32 level)
{
  m_keys.SetAt(index, msgKey);
  m_flags.SetAt(index, flags);
  m_levels.SetAt(index, level);
  nsCOMPtr<nsIMsgFolder> folder;
  hdr->GetFolder(getter_AddRefs(folder));
  m_folders.ReplaceObjectAt(folder, index);
}

PRBool nsMsgSearchDBView::InsertEmptyRows(nsMsgViewIndex viewIndex, PRInt32 numRows)
{
  for (PRInt32 i = 0; i < numRows; i++)
    if (!m_folders.InsertObjectAt(nsnull, viewIndex + i))
      return PR_FALSE;
  return nsMsgDBView::InsertEmptyRows(viewIndex, numRows);
}

void nsMsgSearchDBView::RemoveRows(nsMsgViewIndex viewIndex, PRInt32 numRows)
{
  nsMsgDBView::RemoveRows(viewIndex, numRows);
  for (PRInt32 i = 0; i < numRows; i++)
    m_folders.RemoveObjectAt(viewIndex);
}

nsresult nsMsgSearchDBView::GetMsgHdrForViewIndex(nsMsgViewIndex index, 
                                                  nsIMsgDBHdr **msgHdr)
{
  nsresult rv = NS_MSG_INVALID_DBVIEW_INDEX;
  if (index == nsMsgViewIndex_None || index >= (PRUint32) m_folders.Count())
    return rv;
  nsIMsgFolder *folder = m_folders[index];
  if (folder)
  {
    nsCOMPtr <nsIMsgDatabase> db;
    nsCOMPtr<nsIMsgWindow> msgWindow(do_QueryReferent(mMsgWindowWeak));
    rv = folder->GetMsgDatabase(msgWindow, getter_AddRefs(db));
    NS_ENSURE_SUCCESS(rv, rv);
    if (db)
      rv = db->GetMsgHdrForKey(m_keys[index], msgHdr);
  }
  return rv;
}

NS_IMETHODIMP nsMsgSearchDBView::GetFolderForViewIndex(nsMsgViewIndex index, nsIMsgFolder **aFolder)
{
  NS_ENSURE_ARG_POINTER(aFolder);
  if (index == nsMsgViewIndex_None || index >= (PRUint32) m_folders.Count())
    return NS_MSG_INVALID_DBVIEW_INDEX;
  NS_IF_ADDREF(*aFolder = m_folders[index]);
  return NS_OK;
}

nsresult nsMsgSearchDBView::GetDBForViewIndex(nsMsgViewIndex index, nsIMsgDatabase **db)
{
  nsCOMPtr <nsIMsgFolder> aFolder;
  nsresult rv = GetFolderForViewIndex(index, getter_AddRefs(aFolder));
  NS_ENSURE_SUCCESS(rv, rv);
  return aFolder->GetMsgDatabase(nsnull, db);
}

nsresult nsMsgSearchDBView::AddHdr(nsIMsgDBHdr *msgHdr, nsMsgViewIndex *resultIndex)
{
  nsCOMPtr <nsIMsgFolder> msgFolder;
  msgHdr->GetFolder(getter_AddRefs(msgFolder));
  return AddHdrFromFolder(msgHdr, msgFolder);
}

#ifndef NO_POSTBOX
nsresult nsMsgSearchDBView::AddEntireFolderThreadToXFThread(nsIMsgDBHdr * aMsgHdr, nsMsgXFViewThread * aXFThread, 
                                                            PRUint32 *aThreadFlags, nsCOMArray<nsIMsgDBHdr>& aHdrsAdded)
{
  // in this case we want the real thread
  nsCOMPtr<nsIMsgThread> realThread;
  nsCOMPtr<nsIMsgDatabase> msgDB;
  nsresult rv = GetDBForHeader(aMsgHdr, getter_AddRefs(msgDB));
  NS_ENSURE_SUCCESS(rv, rv);
  rv = msgDB->GetThreadContainingMsgHdr(aMsgHdr, getter_AddRefs(realThread));
  NS_ENSURE_SUCCESS(rv, rv);
  // now add each hdr in the thread
  PRUint32 numChildren;
  realThread->GetNumChildren(&numChildren);
  nsCOMPtr<nsIMsgDBHdr> msgHdr;
  nsCOMPtr<nsIMsgDBHdr> parent;
  PRUint32 posInThread;
  nsCOMPtr<nsIMsgThread> thread (do_QueryInterface(aXFThread));
  PRUint32 msgFlags = 0;
  for (PRUint32 index = 0; index < numChildren; index++)
  {
    realThread->GetChildAt(index, getter_AddRefs(msgHdr));
    if (msgHdr && aXFThread->HdrIndex(msgHdr) == -1) // make sure we aren't adding a hdr that's already there
    {
      AddMsgToHashTables(msgHdr, thread);
      aXFThread->AddHdr(msgHdr, PR_FALSE, posInThread, getter_AddRefs(parent));
      aHdrsAdded.AppendObject(msgHdr);
      msgHdr->GetFlags(&msgFlags);
      *aThreadFlags |= msgFlags & (MSG_FLAG_MARKED | MSG_FLAG_PENDING | MSG_FLAG_ATTACHMENT); // any flags that propogate to the the root
    }
  }
  return NS_OK;
}
#endif

nsresult nsMsgSearchDBView::AddHdrFromFolder(nsIMsgDBHdr *msgHdr, nsIMsgFolder *folder)
{
  if (m_viewFlags & nsMsgViewFlagsType::kGroupBySort)
    return nsMsgGroupView::OnNewHeader(msgHdr, nsMsgKey_None, PR_TRUE);

  PRBool autoSelect = PR_FALSE;
  if (!m_editingTaskMessageID.IsEmpty())
  {
    nsresult rv = ShouldAutoSelectHeader(msgHdr, &autoSelect);
    NS_ENSURE_SUCCESS(rv, rv);
  }

  nsMsgKey msgKey;
  PRUint32 msgFlags;
  msgHdr->GetMessageKey(&msgKey);
  msgHdr->GetFlags(&msgFlags);

  if (m_viewFlags & nsMsgViewFlagsType::kThreadedDisplay)
  {
    nsCOMPtr<nsIMsgThread> thread;
    nsCOMPtr<nsIMsgDBHdr> threadRoot;
    // if we find an xf thread in the hash table corresponding to the new msg's
    // message id, a previous header must be a reference child of the new 
    // message, which means we need to reparent later.
    PRBool msgIsReferredTo;
    GetXFThreadFromMsgHdr(msgHdr, getter_AddRefs(thread), &msgIsReferredTo);
    PRBool newThread = !thread;
    nsMsgXFViewThread *viewThread;
    if (!thread)
    {
      viewThread = new nsMsgXFViewThread(this);
      if (!viewThread)
        return NS_ERROR_OUT_OF_MEMORY;
      thread = do_QueryInterface(viewThread);
    }
    else
    {
      viewThread = static_cast<nsMsgXFViewThread*>(thread.get());
#ifdef NO_POSTBOX
      thread->GetChildAt(0, getter_AddRefs(threadRoot));
#else
      if (viewThread->HdrIndex(msgHdr) != -1)
        return NS_OK;
      GetThreadRootHdr(thread, getter_AddRefs(threadRoot));
#endif
    }

    nsMsgViewIndex threadIndex = nsMsgViewIndex_None;
    if (!newThread && viewThread->MsgCount())
      threadIndex = GetIndexOfFirstDisplayedKeyInThreadUsingHdr(threadRoot);
    PRUint32 newThreadFlags = 0;
    nsCOMArray<nsIMsgDBHdr> addedHdrs;
    AddEntireFolderThreadToXFThread(msgHdr, viewThread, &newThreadFlags, addedHdrs);
    if (newThread || !threadRoot /* a short cut for saying the XF thread we found was empty */)
    {
      GetThreadRootHdr(thread, getter_AddRefs(threadRoot));
      if (!threadRoot)
        return NS_ERROR_FAILURE;
      nsMsgViewIndex insertIndex = GetIndexForThread(threadRoot, newThreadFlags);
      NS_ASSERTION(insertIndex == m_levels.GetSize() || !m_levels[insertIndex],
                    "inserting into middle of thread");
      if (insertIndex == nsMsgViewIndex_None)
        return NS_ERROR_FAILURE;
      threadRoot->GetMessageKey(&msgKey);
      threadRoot->GetFlags(&msgFlags);
      // Postbox collapses all new threads due to Bug 6585
#ifdef NO_POSTBOX
      if (!(m_viewFlags & nsMsgViewFlagsType::kExpandAll))
#endif
        msgFlags |= MSG_FLAG_ELIDED;

      // add any rich rows if appropriate
      PRInt32 numRichRowsAdded = InsertRichRowsForHdr(insertIndex, threadRoot, msgKey, msgFlags | newThreadFlags, 0 /* level */);
      if (numRichRowsAdded)
      {
        NoteChange(insertIndex, numRichRowsAdded, nsMsgViewNotificationCode::insertOrDelete);
        insertIndex += numRichRowsAdded;
      }
      InsertMsgHdrAt(insertIndex, threadRoot, msgKey, msgFlags | newThreadFlags, 0);
      if (viewThread->MsgCount() > 1)
        OrExtraFlag(insertIndex, MSG_VIEW_FLAG_HASCHILDREN | MSG_VIEW_FLAG_ISTHREAD | MSG_VIEW_FLAG_DUMMY);
      NoteChange(insertIndex, 1, nsMsgViewNotificationCode::insertOrDelete);
    }
    else // thread already exists
    {
      nsCOMPtr<nsIMsgDBHdr> newThreadRoot;
      GetThreadRootHdr(thread, getter_AddRefs(newThreadRoot));
      if (!newThreadRoot)
        return NS_ERROR_UNEXPECTED;

      if (threadIndex == nsMsgViewIndex_None)
      {
        // see bug #7767 for details
        NS_ERROR("couldn't find thread index");
        return NS_ERROR_UNEXPECTED;
      }

      // if the old root already has the DUMMY and ISTHREAD flag set, save it.
      // if the old root doesn't have these two flags set (i.e. it was a single message and is not a conversation), delay setting them until after MoveThreadAt
      // is called.
      PRUint32 savedThreadFlags = (m_flags[threadIndex] & (MSG_FLAG_MARKED | MSG_FLAG_PENDING | MSG_FLAG_ATTACHMENT | MSG_FLAG_ELIDED | MSG_VIEW_FLAG_ISTHREAD | MSG_VIEW_FLAG_DUMMY));
      if (newThreadRoot != threadRoot)
      {
        newThreadRoot->GetMessageKey(&msgKey);
        newThreadRoot->GetFlags(&msgFlags);
        SetMsgHdrAt(newThreadRoot, threadIndex, msgKey, msgFlags, 0);
        
        // update any rich rows to reflect the new root
        if (threadIndex && (m_viewFlags & nsMsgViewFlagsType::kRichView))
        {
          // FIX ME if we ever add support for more than one rich row
          PRInt32 richRowFlags = savedThreadFlags | msgFlags | newThreadFlags | MSG_VIEW_FLAG_TITLE_ROW;
          richRowFlags  &= ~(MSG_FLAG_ELIDED | MSG_VIEW_FLAG_HASCHILDREN | MSG_VIEW_FLAG_ISTHREAD | MSG_VIEW_FLAG_DUMMY);
          SetMsgHdrAt(newThreadRoot, threadIndex - 1, msgKey, richRowFlags, 0);
        }
      }
      OrExtraFlag(threadIndex, savedThreadFlags | newThreadFlags | MSG_VIEW_FLAG_HASCHILDREN);
      if (!(m_flags[threadIndex] & MSG_FLAG_ELIDED))
      {
        OrExtraFlag(threadIndex, MSG_VIEW_FLAG_ISTHREAD | MSG_VIEW_FLAG_DUMMY);
        NoteChange(threadIndex, 1, nsMsgViewNotificationCode::changed);
        PRUint32 threadCount;
        thread->GetNumChildren(&threadCount);
        // we are inserting into an expanded thread, so insert each hdr we added
        for (PRUint32 i = 0; i < addedHdrs.Count(); i++)
        {
          nsIMsgDBHdr *hdrToInsert = addedHdrs[i];
          nsMsgViewIndex insertIndex = GetInsertInfoForNewHdr(addedHdrs[i], threadIndex, 1, threadIndex, threadCount - addedHdrs.Count() + i);
          nsMsgKey msgKeyToInsert;
          addedHdrs[i]->GetMessageKey(&msgKeyToInsert);
          PRUint32 msgFlagsToInsert;
          addedHdrs[i]->GetFlags(&msgFlagsToInsert);
          InsertMsgHdrAt(insertIndex, addedHdrs[i], msgKeyToInsert, msgFlagsToInsert, 1);
          NoteChange(insertIndex, 1, nsMsgViewNotificationCode::insertOrDelete);
        }
      }
      else
      {
        NoteChange(threadIndex, 1, nsMsgViewNotificationCode::changed);
      }
      // If the hdr being added is marked then we may need to move the thread due to pinning
      if (newThreadRoot != threadRoot || !(savedThreadFlags & MSG_FLAG_MARKED) && newThreadFlags & MSG_FLAG_MARKED)
        threadIndex = MoveThreadAt(newThreadRoot, threadIndex);
      // we avoid adding the DUMMY flag until after MoveThreadAt for a single message that just became a conversations
      // due to weird issue where we end up reloading the message
      // after MoveThreadAt unfreezes selection because of this check:
      // if ((m_flags[aPrimaryViewIndex] & MSG_VIEW_FLAG_DUMMY) != 0) == m_currentlyDisplayedMsgIsDummyRow) 
      // in nsMsgDBView::LoadMessages which expects the DUMMY flag to not be set yet.
      if (threadIndex != nsMsgViewIndex_None)
      {
        OrExtraFlag(threadIndex, MSG_VIEW_FLAG_ISTHREAD | MSG_VIEW_FLAG_DUMMY);
        NoteChange(threadIndex, 1, nsMsgViewNotificationCode::changed);
      }
    }
  }
  else
  {
    // nsMsgKey_None means it's not a valid hdr.
    if (msgKey != nsMsgKey_None)
    {
      msgHdr->GetFlags(&msgFlags);
      m_folders.AppendObject(folder);
      m_keys.Add(msgKey);
      m_levels.Add(0);
      m_flags.Add(msgFlags);
      NoteChange(GetSize() - 1, 1, nsMsgViewNotificationCode::insertOrDelete);
    }
  }

  if (autoSelect)
  {
    nsMsgViewIndex viewIndex = FindHdr(msgHdr, 0, PR_FALSE);
    if (mTree && mTreeSelection && viewIndex != nsMsgViewIndex_None)
    {
      mTreeSelection->Select(viewIndex);
      // make sure the selected message is visible in the thread pane
      // so we don't have to scroll to see it.  
      // if we are sorted by date ascending
      // and we have more messages than visible rows, this prevents our
      // message from being partially visible.
      // see bug #630 for details
      mTree->EnsureRowIsVisible(viewIndex);
    }
    
  }
  return NS_OK;
}

// This method removes the thread at threadIndex from the view 
// and puts it back in its new position, determined by the sort order.
// And, if the selection is affected, save and restore the selection.
nsMsgViewIndex nsMsgSearchDBView::MoveThreadAt(nsIMsgDBHdr *threadRoot, nsMsgViewIndex threadIndex)
{
  PRBool updatesSuppressed = mSuppressChangeNotification;
  // Turn off tree notifications so that we don't reload the current message.
  if (!updatesSuppressed)
    DisableChangeUpdates();

  nsCOMPtr<nsIMsgDBHdr> threadHdr;
  GetMsgHdrForViewIndex(threadIndex, getter_AddRefs(threadHdr));
  if (!threadHdr)
    return nsMsgViewIndex_None;

  PRUint32 saveFlags = m_flags[threadIndex];
  PRBool threadIsExpanded = !(saveFlags & MSG_FLAG_ELIDED);
  PRInt32 childCount = 0;
  nsMsgKey preservedKey;
  nsMsgKeyArray preservedSelection;
  PRInt32 selectionCount;
  PRInt32 currentIndex;
  PRBool hasSelection = mTree && mTreeSelection &&
                        ((NS_SUCCEEDED(mTreeSelection->GetCurrentIndex(&currentIndex)) &&
                         currentIndex >= 0 && currentIndex < GetSize()) ||
                         (NS_SUCCEEDED(mTreeSelection->GetRangeCount(&selectionCount)) &&
                          selectionCount > 0));


  if (hasSelection)
    SaveAndClearSelection(&preservedKey, &preservedSelection);

  if (threadIsExpanded)
  {
    ExpansionDelta(threadIndex, &childCount);
    childCount = -childCount;
  }
  nsMsgKeyArray threadKeys;
  nsUInt32Array threadFlags;
  nsUint8Array threadLevels;
  nsCOMArray<nsIMsgFolder> threadFolders;
  nsCOMPtr<nsIMsgThread> thread;
  GetThreadContainingMsgHdr(threadRoot, getter_AddRefs(thread));
  if (threadIsExpanded)
  {
#ifdef NO_POSTBOX
    threadKeys.SetSize(childCount);
    threadFlags.SetSize(childCount);
    threadLevels.SetSize(childCount);
#endif
    threadFolders.SetCapacity(childCount);
    for (nsMsgViewIndex index = threadIndex + 1; 
        index < (nsMsgViewIndex) GetSize() && m_levels[index]; index++)
    {
      threadKeys.Add(m_keys[index]);
      threadFlags.Add(m_flags[index]);
      threadLevels.Add(m_levels[index]);
      threadFolders.AppendObject(m_folders[index]);
    }
    PRUint32 collapseCount;
    CollapseByIndex(threadIndex, &collapseCount);
  }
  nsMsgDBView::RemoveByIndex(threadIndex);
  m_folders.RemoveObjectAt(threadIndex);
#ifndef NO_POSTBOX
  // Certain fields such as MSG_FLAG_MARKED roll up to the top of the conversation if any child has that flag.
  // adjust the fields for the thread root index to include this
  saveFlags |= GetFlagsForThread(thread, MSG_FLAG_MARKED | MSG_FLAG_PENDING | MSG_FLAG_ATTACHMENT);
#endif

  nsMsgKey richListRowKey;
  if (threadIndex && (m_viewFlags & nsMsgViewFlagsType::kRichView) && IsRichRow(threadIndex - 1))
  {
    // FIX ME if we add support for more than one rich list row
    richListRowKey = m_keys[threadIndex - 1];
    nsMsgDBView::RemoveByIndex(threadIndex - 1);
    m_folders.RemoveObjectAt(threadIndex - 1);
  }

  nsMsgViewIndex newIndex = GetIndexForThread(threadHdr, saveFlags);
  NS_ASSERTION(newIndex == m_levels.GetSize() || !m_levels[newIndex],
                "inserting into middle of thread");
  if (newIndex == nsMsgViewIndex_None)
    newIndex = 0;

  nsMsgKey msgKey;
  PRUint32 msgFlags;
  threadHdr->GetMessageKey(&msgKey);
  threadHdr->GetFlags(&msgFlags);

  // move any rich list rows first
  if ((m_viewFlags & nsMsgViewFlagsType::kRichView) && richListRowKey != nsMsgKey_None)
    newIndex += InsertRichRowsForHdr(newIndex, threadHdr, msgKey, saveFlags, 0);

  InsertMsgHdrAt(newIndex, threadHdr, msgKey, msgFlags, 0);

  if (threadIsExpanded)
  {
    m_keys.InsertAt(newIndex + 1, &threadKeys);
    m_flags.InsertAt(newIndex + 1, &threadFlags);
    m_levels.InsertAt(newIndex + 1, &threadLevels);
    m_folders.InsertObjectsAt(threadFolders, newIndex + 1);
  }
  m_flags[newIndex] = saveFlags;
  // unfreeze selection.
  if (hasSelection)
    RestoreSelection(preservedKey, &preservedSelection);

  if (!updatesSuppressed)
    EnableChangeUpdates();
  nsMsgViewIndex lowIndex = threadIndex < newIndex ? threadIndex : newIndex;
  nsMsgViewIndex highIndex = lowIndex == threadIndex ? newIndex : threadIndex;
  NoteChange(lowIndex, highIndex - lowIndex + childCount + 1,
             nsMsgViewNotificationCode::changed);
  return newIndex;
}

nsresult nsMsgSearchDBView::InsertHdrFromFolder(nsIMsgDBHdr *msgHdr, nsIMsgFolder *folder)
{
  nsMsgViewIndex insertIndex = nsMsgViewIndex_None;
  // Threaded view always needs to go through AddHdrFromFolder since
  // it handles the xf view thread object creation.
  if (! (m_viewFlags & nsMsgViewFlagsType::kThreadedDisplay))
    insertIndex = GetInsertIndex(msgHdr, 0);

  if (insertIndex == nsMsgViewIndex_None)
    return AddHdrFromFolder(msgHdr, folder);

  nsMsgKey msgKey;
  PRUint32 msgFlags;
  msgHdr->GetMessageKey(&msgKey);
  msgHdr->GetFlags(&msgFlags);
  
  PRInt32 numRichRowsInserted = InsertRichRowsForHdr(insertIndex, msgHdr, msgKey, msgFlags, 0);
  if (numRichRowsInserted)
  {
    NoteChange(insertIndex, numRichRowsInserted, nsMsgViewNotificationCode::insertOrDelete);
    insertIndex += numRichRowsInserted;
  }

  InsertMsgHdrAt(insertIndex, msgHdr, msgKey, msgFlags, 0);

  // the call to NoteChange() has to happen after we add the key
  // as NoteChange() will call RowCountChanged() which will call our GetRowCount()
  NoteChange(insertIndex, 1, nsMsgViewNotificationCode::insertOrDelete);
  return NS_OK;
}

NS_IMETHODIMP
nsMsgSearchDBView::OnSearchHit(nsIMsgDBHdr* aMsgHdr, nsIMsgFolder *folder)
{
  NS_ENSURE_ARG(aMsgHdr);
  NS_ENSURE_ARG(folder);

  nsCOMPtr <nsISupports> supports = do_QueryInterface(folder);
  if (m_folders.IndexOf(folder) < 0 ) //do this just for new folder
  {
    nsCOMPtr<nsIMsgDatabase> dbToUse;
    nsCOMPtr<nsIDBFolderInfo> folderInfo;
    folder->GetDBFolderInfoAndDB(getter_AddRefs(folderInfo), getter_AddRefs(dbToUse));
    if (dbToUse)
    {
      dbToUse->AddListener(this);
      m_dbToUseList.AppendObject(dbToUse);
    }
  }

  if (m_sortValid)
    return InsertHdrFromFolder(aMsgHdr, folder);
  else
    return AddHdrFromFolder(aMsgHdr, folder);
}

NS_IMETHODIMP
nsMsgSearchDBView::OnSearchDone(nsresult status)
{
  //we want to set imap delete model once the search is over because setting next
  //message after deletion will happen before deleting the message and search scope
  //can change with every search.
  mDeleteModel = nsMsgImapDeleteModels::MoveToTrash;  //set to default in case it is non-imap folder
  nsIMsgFolder *curFolder = m_folders.SafeObjectAt(0);
  if (curFolder)   
    GetImapDeleteModel(curFolder);
  return NS_OK;
}

// for now also acts as a way of resetting the search datasource
NS_IMETHODIMP
nsMsgSearchDBView::OnNewSearch()
{
  PRInt32 oldSize = GetSize();

  PRInt32 count = m_dbToUseList.Count();
  for(PRInt32 j = 0; j < count; j++) 
    m_dbToUseList[j]->RemoveListener(this);

  m_dbToUseList.Clear();

  m_folders.Clear();
  m_keys.RemoveAll();
  m_levels.RemoveAll();
  m_flags.RemoveAll();

  // needs to happen after we remove the keys, since RowCountChanged() will call our GetRowCount()
  if (mTree) 
    mTree->RowCountChanged(0, -oldSize);

//    mSearchResults->Clear();
    return NS_OK;
}

NS_IMETHODIMP nsMsgSearchDBView::OnAnnouncerGoingAway(nsIDBChangeAnnouncer *instigator)
{
  nsIMsgDatabase *db = static_cast<nsIMsgDatabase *>(instigator);
  if (db)
  {
    db->RemoveListener(this);
    m_dbToUseList.RemoveObject(db);
  }
  return NS_OK;
}

nsCOMArray<nsIMsgFolder>* nsMsgSearchDBView::GetFolders()
{
  return &m_folders;
}

NS_IMETHODIMP
nsMsgSearchDBView::GetCommandStatus(nsMsgViewCommandTypeValue command, PRBool *selectable_p, nsMsgViewCommandCheckStateValue *selected_p)
{
  if (command != nsMsgViewCommandType::runJunkControls)
    return nsMsgDBView::GetCommandStatus(command, selectable_p, selected_p);

  *selectable_p = PR_FALSE;
  return NS_OK;
}

NS_IMETHODIMP 
nsMsgSearchDBView::DoCommandWithFolder(nsMsgViewCommandTypeValue command, nsIMsgFolder *destFolder)
{
    mCommand = command;
    mDestFolder = destFolder;
    return nsMsgDBView::DoCommandWithFolder(command, destFolder);
}

NS_IMETHODIMP nsMsgSearchDBView::DoCommand(nsMsgViewCommandTypeValue command)
{
  mCommand = command;
  if (command == nsMsgViewCommandType::deleteMsg || 
      command == nsMsgViewCommandType::archive ||
      command == nsMsgViewCommandType::rename ||
      command == nsMsgViewCommandType::deleteNoTrash ||
      command == nsMsgViewCommandType::selectAll || 
      command ==nsMsgViewCommandType::expandAll ||
      command == nsMsgViewCommandType::collapseAll)
    return nsMsgDBView::DoCommand(command);
  nsresult rv = NS_OK;

  nsUInt32Array selection;
  if (command == nsMsgViewCommandType::markAllRead)
  {
    for (PRInt32 i = 0; i < GetSize(); i++)
      selection.Add(i);
    command = nsMsgViewCommandType::markMessagesRead;
  }
  else
  {
    GetSelectedIndices(&selection);
  }

  nsMsgViewIndex *indices = selection.GetData();
  PRInt32 numIndices = selection.GetSize();

  // we need to break apart the selection by folders, and then call
  // ApplyCommandToIndices with the command and the indices in the
  // selection that are from that folder.

  nsUInt32Array *indexArrays;
  PRInt32 numArrays;
  rv = PartitionSelectionByFolder(indices, numIndices, &indexArrays, &numArrays);
  NS_ENSURE_SUCCESS(rv, rv);
  for (PRInt32 folderIndex = 0; folderIndex < numArrays; folderIndex++)
  {
    rv = ApplyCommandToIndices(command, indexArrays[folderIndex].GetData(), indexArrays[folderIndex].GetSize());
    NS_ENSURE_SUCCESS(rv, rv);
  }

  return rv;
}

// This method removes the specified line from the view, and adjusts the
// various flags and levels of affected messages.
nsresult nsMsgSearchDBView::RemoveByIndex(nsMsgViewIndex index)
{
    if (!IsValidIndex(index))
        return NS_MSG_INVALID_DBVIEW_INDEX;

  if (m_viewFlags & nsMsgViewFlagsType::kThreadedDisplay)
  {
    nsCOMPtr<nsIMsgDBHdr> msgHdr;
    nsCOMPtr<nsIMsgThread> thread;
    nsresult rv = GetMsgHdrForViewIndex(index, getter_AddRefs(msgHdr));
    NS_ENSURE_SUCCESS(rv, rv);
    
    GetXFThreadFromMsgHdr(msgHdr, getter_AddRefs(thread));
    if (thread)
    {
      nsMsgXFViewThread *viewThread = static_cast<nsMsgXFViewThread*>(thread.get());
      if (viewThread->MsgCount() == 2)
      {
        // if we removed the next to last message in the thread,
        // we need to adjust the flags on the first message in the thread.
        nsMsgViewIndex threadIndex = m_levels[index] ? index -1 : index;
        if (threadIndex != nsMsgViewIndex_None)
        {
          AndExtraFlag(threadIndex, ~(MSG_VIEW_FLAG_ISTHREAD | MSG_FLAG_ELIDED |
                                      MSG_VIEW_FLAG_HASCHILDREN));
          m_levels[threadIndex] = 0;
          NoteChange(threadIndex, 1, nsMsgViewNotificationCode::changed);
        }
      }
#ifdef NO_POSTBOX
      // Bump up the level of all the descendents of the message
      // that was removed, if the thread was expanded.
      PRUint8 removedLevel = m_levels[index];
      nsMsgViewIndex i = index + 1;
      if (i < m_levels.GetSize() && m_levels[i] > removedLevel)
      {
        // promote the child of the removed message.
        PRUint8 promotedLevel = m_levels[i];
        m_levels[i] = promotedLevel - 1;
        i++;
        // now promote all the children of the promoted message.
        for (; i < m_levels.GetSize() && 
              m_levels[i] > promotedLevel; i++)
          m_levels[i] = m_levels[i] - 1;
      }
#endif
    }
  }
  m_folders.RemoveObjectAt(index);
  return nsMsgDBView::RemoveByIndex(index);
}

nsresult nsMsgSearchDBView::ArchiveMessages(nsIMsgWindow *window, nsMsgViewIndex *indices, PRInt32 numIndices)
{
   GetFoldersAndHdrsForSelection(indices, numIndices);

   // remember the archived messages in case the user undoes the archive,
   // and we want to restore the hdr to the view, even if it no
   // longer matches the search criteria.
    for (nsMsgViewIndex i = 0; i < (nsMsgViewIndex) numIndices; i++) 
    {
      nsCOMPtr<nsIMsgDBHdr> msgHdr; 
      (void) GetMsgHdrForViewIndex(indices[i],getter_AddRefs(msgHdr));
      if (msgHdr)
        RememberDeletedMsgHdr(msgHdr);
    }
    return ProcessRequestsInOneFolder(window);
}

nsresult nsMsgSearchDBView::DeleteMessages(nsIMsgWindow *window, nsMsgViewIndex *indices, PRInt32 numIndices, PRBool deleteStorage)
{
   nsresult rv = GetFoldersAndHdrsForSelection(indices, numIndices);
   NS_ENSURE_SUCCESS(rv, rv);
   if (mDeleteModel != nsMsgImapDeleteModels::MoveToTrash)
     deleteStorage = PR_TRUE;
  if (mDeleteModel != nsMsgImapDeleteModels::IMAPDelete)
    m_deletingRows = PR_TRUE;

  // remember the deleted messages in case the user undoes the delete,
  // and we want to restore the hdr to the view, even if it no
  // longer matches the search criteria.
  for (nsMsgViewIndex i = 0; i < (nsMsgViewIndex) numIndices; i++) 
  {
    nsCOMPtr<nsIMsgDBHdr> msgHdr; 
    (void) GetMsgHdrForViewIndex(indices[i], getter_AddRefs(msgHdr));
    if (msgHdr)
      RememberDeletedMsgHdr(msgHdr);
    // if we are deleting rows, save off the view indices
    if (m_deletingRows)
      mIndicesToNoteChange.Add(indices[i]);

  }
  rv = deleteStorage ? ProcessRequestsInAllFolders(window)
                     : ProcessRequestsInOneFolder(window);
  if (NS_FAILED(rv))
    m_deletingRows = PR_FALSE;
  return rv;
}

nsresult 
nsMsgSearchDBView::CopyMessages(nsIMsgWindow *window, nsMsgViewIndex *indices, PRInt32 numIndices, PRBool isMove, nsIMsgFolder *destFolder)
{
    GetFoldersAndHdrsForSelection(indices, numIndices);

    return ProcessRequestsInOneFolder(window);
}

nsresult
nsMsgSearchDBView::PartitionSelectionByFolder(nsMsgViewIndex *indices, PRInt32 numIndices, nsUInt32Array **indexArrays, PRInt32 *numArrays)
{
  nsMsgViewIndex i;
  PRInt32 folderIndex;
  nsCOMArray<nsIMsgFolder> uniqueFoldersSelected;
  nsTArray<PRUint32> numIndicesSelected;
  mCurIndex = 0;

  //Build unique folder list based on headers selected by the user
  for (i = 0; i < (nsMsgViewIndex) numIndices; i++)
  {
    nsIMsgFolder *curFolder = m_folders[indices[i]];
    folderIndex = uniqueFoldersSelected.IndexOf(curFolder);
    if (folderIndex < 0)
    {
      uniqueFoldersSelected.AppendObject(curFolder);
      numIndicesSelected.AppendElement(1);
    }
    else
    {
      numIndicesSelected[folderIndex]++;
    }
  }

  PRUint32 numFolders = uniqueFoldersSelected.Count();   //group the headers selected by each folder 
  *indexArrays = new nsUInt32Array[numFolders];
  *numArrays = numFolders;
  NS_ENSURE_TRUE(*indexArrays, NS_ERROR_OUT_OF_MEMORY);
  for (PRUint32 folderIndex=0; folderIndex < numFolders; folderIndex++)
  {
     for (nsMsgViewIndex i = 0; i < (nsMsgViewIndex) numIndices; i++) 
     {
       nsIMsgFolder *curFolder = m_folders[indices[i]];
       PRInt32 folderIndex = uniqueFoldersSelected.IndexOf(curFolder);
       (*indexArrays)[folderIndex].Add(indices[i]);
     }
  }
  return NS_OK;
}

nsresult
nsMsgSearchDBView::GetFoldersAndHdrsForSelection(nsMsgViewIndex *indices, PRInt32 numIndices)
{
  nsresult rv = NS_OK; 
  mCurIndex = 0;
  m_uniqueFoldersSelected.Clear();
  
  if (!m_hdrsForEachFolder)
  {
    m_hdrsForEachFolder = do_CreateInstance(NS_SUPPORTSARRAY_CONTRACTID, &rv);
    NS_ENSURE_SUCCESS(rv,rv);
  }
  else
    m_hdrsForEachFolder->Clear();

#ifndef NO_POSTBOX
  // account for actions applying to all hdrs in a conversation
  nsCOMPtr<nsISupportsArray> messages;
  PRBool hasConversations;
  rv = BuildUpMessagesFromSelection(indices, numIndices, &hasConversations, getter_AddRefs(messages));
  NS_ENSURE_SUCCESS(rv, rv);
  PRUint32 numMsgs;
  messages->Count(&numMsgs);
#endif

  // Build unique folder list based on headers selected by the user
  PRUint32 i;
  for (i = 0; i < numMsgs; i++)
  {
    nsCOMPtr<nsIMsgDBHdr> hdr = do_QueryElementAt(messages, i, &rv);
    if (hdr)
    {
      nsCOMPtr<nsIMsgFolder> curFolder;
      hdr->GetFolder(getter_AddRefs(curFolder));
      if (m_uniqueFoldersSelected.IndexOf(curFolder) < 0)
        m_uniqueFoldersSelected.AppendObject(curFolder);
    }
  }

  // Group the headers selected by each folder
  PRUint32 numFolders = m_uniqueFoldersSelected.Count();
  for (PRUint32 folderIndex = 0; folderIndex < numFolders; folderIndex++)
  {
     nsIMsgFolder *curFolder = m_uniqueFoldersSelected[folderIndex];
#ifdef NO_POSTBOX
     nsCOMPtr<nsIMutableArray> msgHdrsForOneFolder(do_CreateInstance(NS_ARRAY_CONTRACTID));
#else
     nsCOMPtr <nsISupportsArray> msgHdrsForOneFolder;
     NS_NewISupportsArray(getter_AddRefs(msgHdrsForOneFolder));     
#endif
     for (i = 0; i < numMsgs; i++)
     {
        nsCOMPtr<nsIMsgDBHdr> hdr = do_QueryElementAt(messages, i, &rv);
        if (hdr)
        {
          nsCOMPtr<nsIMsgFolder> msgFolder;
          hdr->GetFolder(getter_AddRefs(msgFolder));
          if (NS_SUCCEEDED(rv) && msgFolder && msgFolder == curFolder) 
          {
            nsCOMPtr<nsISupports> hdrSupports = do_QueryInterface(hdr);
#ifdef NO_POSTBOX
            msgHdrsForOneFolder->AppendElement(hdrSupports, PR_FALSE);
#else
            msgHdrsForOneFolder->AppendElement(hdrSupports);
#endif
          }
        }
     }
     nsCOMPtr<nsISupports> supports = do_QueryInterface(msgHdrsForOneFolder, &rv);
     if (NS_SUCCEEDED(rv) && supports)
       m_hdrsForEachFolder->AppendElement(supports);
  }
  return rv;
}

nsresult
nsMsgSearchDBView::ApplyCommandToIndicesWithFolder(nsMsgViewCommandTypeValue command, nsMsgViewIndex* indices,
                    PRInt32 numIndices, nsIMsgFolder *destFolder)
{
  mCommand = command;
  mDestFolder = destFolder;
  return nsMsgDBView::ApplyCommandToIndicesWithFolder(command, indices, numIndices, destFolder);
}

// nsIMsgCopyServiceListener methods

NS_IMETHODIMP
nsMsgSearchDBView::OnStopCopy(nsresult aStatus)
{
  if (NS_SUCCEEDED(aStatus))
  {
    mCurIndex++;
    PRUint32 numFolders = 0;
    if (mCurIndex < (PRUint32) m_uniqueFoldersSelected.Count())
    {
      nsCOMPtr<nsIMsgWindow> msgWindow(do_QueryReferent(mMsgWindowWeak));
      ProcessRequestsInOneFolder(msgWindow);
    }
    else
      nsMsgDBView::OnStopCopy(aStatus);
  }
  return NS_OK;
}

// end nsIMsgCopyServiceListener methods

nsresult nsMsgSearchDBView::ProcessRequestsInOneFolder(nsIMsgWindow *window)
{
    nsresult rv = NS_OK;

    // Folder operations like copy/move are not implemented for .eml files.
    if (m_uniqueFoldersSelected.Count() == 0)
      return NS_ERROR_NOT_IMPLEMENTED;
  
    nsIMsgFolder *curFolder = m_uniqueFoldersSelected[mCurIndex];
    NS_ASSERTION(curFolder, "curFolder is null");
    nsCOMPtr <nsISupportsArray> messageArray =
        do_QueryElementAt(m_hdrsForEachFolder, mCurIndex);
    NS_ASSERTION(messageArray, "messageArray is null");

    if (mCommand == nsMsgViewCommandType::deleteMsg)
      curFolder->DeleteMessages(messageArray, window, PR_FALSE /* delete storage */, PR_FALSE /* is move*/, this, PR_TRUE /*allowUndo*/);
    else if (mCommand == nsMsgViewCommandType::archive)
      ArchiveMessagesFromFolder(window, curFolder, messageArray);
    else 
    {
      NS_ASSERTION(!(curFolder == mDestFolder), "The source folder and the destination folder are the same");
      if (NS_SUCCEEDED(rv) && curFolder != mDestFolder)
      {
         nsCOMPtr<nsIMsgCopyService> copyService = do_GetService(NS_MSGCOPYSERVICE_CONTRACTID, &rv);
         if (NS_SUCCEEDED(rv))
         {
           if (mCommand == nsMsgViewCommandType::moveMessages)
             copyService->CopyMessages(curFolder, messageArray, mDestFolder, PR_TRUE /* isMove */, this, window, PR_TRUE /*allowUndo*/);
           else if (mCommand == nsMsgViewCommandType::copyMessages)
             copyService->CopyMessages(curFolder, messageArray, mDestFolder, PR_FALSE /* isMove */, this, window, PR_TRUE /*allowUndo*/);
         }
      }
    }
    return rv;
}

nsresult nsMsgSearchDBView::ProcessRequestsInAllFolders(nsIMsgWindow *window)
{
  PRUint32 numFolders = m_uniqueFoldersSelected.Count();
  for (PRUint32 folderIndex = 0; folderIndex < numFolders; folderIndex++)
  {
    nsIMsgFolder *curFolder = m_uniqueFoldersSelected[folderIndex];
    NS_ASSERTION (curFolder, "curFolder is null");

     nsCOMPtr <nsISupportsArray> messageArray =
         do_QueryElementAt(m_hdrsForEachFolder, folderIndex);
     NS_ASSERTION(messageArray, "messageArray is null");

     curFolder->DeleteMessages(messageArray, window, PR_TRUE /* delete storage */, PR_FALSE /* is move*/, nsnull/*copyServListener*/, PR_FALSE /*allowUndo*/ );
  }

  return NS_OK;
}

NS_IMETHODIMP nsMsgSearchDBView::Sort(nsMsgViewSortTypeValue sortType, nsMsgViewSortOrderValue sortOrder)
{
    PRInt32 rowCountBeforeSort = GetSize();

    if (!rowCountBeforeSort)
        return NS_OK;

    if (m_viewFlags & (nsMsgViewFlagsType::kThreadedDisplay |
                      nsMsgViewFlagsType::kGroupBySort))
    {
      // ### This forgets which threads were expanded, and is sub-optimal
      // since it rebuilds the thread objects.  
#ifdef NO_POSTBOX
      m_sortType = sortType;
      m_sortOrder = sortOrder;
#else
      pbAdjustSortFieldsForPinning(sortType, sortOrder);
#endif
      return RebuildView(m_viewFlags);
    }

    nsMsgKey preservedKey;
    nsMsgKeyArray preservedSelection;
    SaveAndClearSelection(&preservedKey, &preservedSelection);

    m_sortType = nsMsgViewSortType::byNone; // sort from scratch
    nsresult rv = nsMsgDBView::Sort(sortType,sortOrder);

    // the sort may have changed the number of rows
    // before we restore the selection, tell the tree
    // do this before we call restore selection
    // this is safe when there is no selection. 
    rv = AdjustRowCount(rowCountBeforeSort, GetSize());

    RestoreSelection(preservedKey, &preservedSelection);
    if (mTree) mTree->Invalidate();

    NS_ENSURE_SUCCESS(rv,rv);
    return rv;
}

nsMsgViewIndex nsMsgSearchDBView::FindHdr(nsIMsgDBHdr *msgHdr, nsMsgViewIndex startIndex,
                                          PRBool allowDummy)
{
  nsCOMPtr<nsIMsgDBHdr> curHdr;
  PRInt32 index;
  // it would be nice to take advantage of sorted views when possible.
  for (index = startIndex; index < GetSize(); index++)
  {
    GetMsgHdrForViewIndex(index, getter_AddRefs(curHdr));
    // never return a rich row
    if (curHdr == msgHdr && !IsRichRow(index) && 
        (allowDummy ||
         !(m_flags[index] & MSG_VIEW_FLAG_DUMMY) ||
         (m_flags[index] & MSG_FLAG_ELIDED)))
      break;
  }
  return index < GetSize() ? index : nsMsgViewIndex_None;
}

// This method looks for the XF thread that corresponds to this message hdr,
// first by looking up the message id, then references, and finally, if subject
// threading is turned on, the subject.
nsresult nsMsgSearchDBView::GetXFThreadFromMsgHdr(nsIMsgDBHdr *msgHdr, 
                                                  nsIMsgThread **pThread,
                                                  PRBool *foundByMessageId)
{
  nsAutoString hashKey;
  nsCAutoString messageId;
  msgHdr->GetMessageId(getter_Copies(messageId));
  CopyASCIItoUTF16(messageId, hashKey);
  *pThread = nsnull;
  m_threadsTable.Get(hashKey, pThread);
  // The caller may want to know if we found the thread by the msgHdr's
  // messageId
  if (foundByMessageId)
    *foundByMessageId = *pThread != nsnull;
  if (!*pThread)
  {
    PRUint16 numReferences = 0;
    msgHdr->GetNumReferences(&numReferences);
    for (PRInt32 i = numReferences - 1; i >= 0  && !*pThread; i--)
    {
      nsCAutoString reference;
      
      msgHdr->GetStringReference(i, reference);
      if (reference.IsEmpty())
        break;

      CopyASCIItoUTF16(reference, hashKey);
      m_threadsTable.Get(hashKey, pThread);
    }
#ifndef NO_POSTBOX
    // Postbox never threads if a message has a reference header but we can't find the thread.
    if (numReferences)
      return (*pThread) ? NS_OK : NS_ERROR_FAILURE;
#endif
  }
  // if we're threading by subject, and we couldn't find the thread by ref,
  // just treat subject as an other ref.
  if (!*pThread && !gReferenceOnlyThreading)
  {
    nsCString subject;
    msgHdr->GetSubject(getter_Copies(subject));

#ifndef NO_POSTBOX
    // respect the thread with RE pref
    if (!gThreadWithoutRe)
    {
      PRUint32 flags;
      msgHdr->GetFlags(&flags);
      if (!(flags & MSG_FLAG_HAS_RE)) // subject does not have Re:
        return NS_ERROR_FAILURE; // no thread found
    }
#endif
    // this is the raw rfc822 subject header, so this is OK
    CopyASCIItoUTF16(subject, hashKey);
    m_threadsTable.Get(hashKey, pThread);
  }
  return (*pThread) ? NS_OK : NS_ERROR_FAILURE;
}

nsresult nsMsgSearchDBView::GetMsgHdrFromHash(nsCString &reference, nsIMsgDBHdr **hdr)
{
  nsString hashKey;
  CopyASCIItoUTF16(reference, hashKey);
  return m_hdrsTable.Get(hashKey, hdr);
}

nsresult nsMsgSearchDBView::GetThreadFromHash(nsCString &reference, 
                                              nsIMsgThread **thread)
{
  nsString hashKey;
  CopyASCIItoUTF16(reference, hashKey);
  return m_threadsTable.Get(hashKey, thread);
}

nsresult nsMsgSearchDBView::AddRefToHash(nsCString &reference, 
                                         nsIMsgThread *thread)
{
  nsString hashKey;
  CopyASCIItoUTF16(reference, hashKey);
  // Check if this reference is already is associated with a thread;
  // If so, don't overwrite that association.
  nsCOMPtr<nsIMsgThread> oldThread;
  m_threadsTable.Get(hashKey, getter_AddRefs(oldThread));
  if (oldThread)
    return NS_OK;

  return m_threadsTable.Put(hashKey, thread);
}

nsresult nsMsgSearchDBView::AddMsgToHashTables(nsIMsgDBHdr *msgHdr,
                                               nsIMsgThread *thread)
{
  PRUint16 numReferences = 0;
  nsresult rv;

  msgHdr->GetNumReferences(&numReferences);
  for (PRInt32 i = 0; i < numReferences; i++)
  {
    nsCAutoString reference;
    
    msgHdr->GetStringReference(i, reference);
    if (reference.IsEmpty())
      break;

    rv = AddRefToHash(reference, thread);
    NS_ENSURE_SUCCESS(rv, rv);
  }

  nsCString messageId;
  msgHdr->GetMessageId(getter_Copies(messageId));
  nsString hashKey;
  CopyASCIItoUTF16(messageId, hashKey);
  m_hdrsTable.Put(hashKey, msgHdr);
  if (!gReferenceOnlyThreading)
  {
    nsCString subject;
    msgHdr->GetSubject(getter_Copies(subject));
    // if we're threading by subject, just treat subject as an other ref.
    AddRefToHash(subject, thread);
  }
  return AddRefToHash(messageId, thread);
}

nsresult nsMsgSearchDBView::RemoveRefFromHash(nsCString &reference)
{
  nsString hashKey;
  CopyASCIItoUTF16(reference, hashKey);
  m_threadsTable.Remove(hashKey);
  return NS_OK;
}

nsresult nsMsgSearchDBView::RemoveMsgFromHashTables(nsIMsgDBHdr *msgHdr)
{
  PRUint16 numReferences = 0;
  nsresult rv = NS_OK;

  msgHdr->GetNumReferences(&numReferences);

  for (PRInt32 i = 0; i < numReferences; i++)
  {
    nsCAutoString reference;
    msgHdr->GetStringReference(i, reference);
    if (reference.IsEmpty())
      break;

    rv = RemoveRefFromHash(reference);
    if (NS_FAILED(rv))
      break;
  }
  nsCString messageId;
  msgHdr->GetMessageId(getter_Copies(messageId));
  nsString hashKey;
  CopyASCIItoUTF16(messageId, hashKey);
  m_hdrsTable.Remove(hashKey);
  RemoveRefFromHash(messageId);
  if (!gReferenceOnlyThreading)
  {
    nsCString subject;
    msgHdr->GetSubject(getter_Copies(subject));
    // if we're threading by subject, just treat subject as an other ref.
    RemoveRefFromHash(subject);
  }
  return rv;
}

nsMsgGroupThread *nsMsgSearchDBView::CreateGroupThread(nsIMsgDatabase * /* db */)
{
  return new nsMsgXFGroupThread();
}

nsresult nsMsgSearchDBView::GetThreadContainingMsgHdr(nsIMsgDBHdr *msgHdr, 
                                                      nsIMsgThread **pThread)
{
  if (m_viewFlags & nsMsgViewFlagsType::kGroupBySort)
    return nsMsgGroupView::GetThreadContainingMsgHdr(msgHdr, pThread);
  else if (m_viewFlags & nsMsgViewFlagsType::kThreadedDisplay)
    return GetXFThreadFromMsgHdr(msgHdr, pThread);

  // if not threaded, use the real thread. 
  nsCOMPtr<nsIMsgDatabase> msgDB;
  nsresult rv = GetDBForHeader(msgHdr, getter_AddRefs(msgDB));
  NS_ENSURE_SUCCESS(rv, rv);
  return msgDB->GetThreadContainingMsgHdr(msgHdr, pThread);
}

#ifndef NO_POSTBOX
struct GetAllKeysEnumStruc
{
  nsMsgKey threadKey;
  nsIMsgThread *thread;
};

nsresult nsMsgSearchDBView::GetThreadContainingThreadKey(nsMsgKey aThreadKey, nsIMsgThread **aThread)
{
  // brute force enumerate the group table looking for one with our thread key
  GetAllKeysEnumStruc struc;
  struc.threadKey = aThreadKey;
  struc.thread = nsnull;
  m_threadsTable.EnumerateRead(GroupThreadKeyEnumerator, (void *) &struc);
  NS_IF_ADDREF(*aThread = struc.thread);
  return NS_OK;
}

PLDHashOperator
nsMsgSearchDBView::GroupThreadKeyEnumerator(const nsAString &aKey, nsIMsgThread* aGroupThread, void* aArg)
{
  GetAllKeysEnumStruc* struc = (GetAllKeysEnumStruc*)aArg;
  nsMsgKey threadKey;
  aGroupThread->GetThreadKey(&threadKey);
  if (threadKey == struc->threadKey)
  {
    struc->thread = aGroupThread;
    return PL_DHASH_STOP;
  }
  return PL_DHASH_NEXT;
}
#endif

nsresult 
nsMsgSearchDBView::ListIdsInThread(nsIMsgThread *threadHdr, 
                                   nsMsgViewIndex startOfThreadViewIndex, 
                                   PRUint32 *pNumListed)
{
  NS_ENSURE_ARG(threadHdr);
  // these children ids should be in thread order.
  PRUint32 i;
  nsMsgViewIndex viewIndex = startOfThreadViewIndex + 1;
  *pNumListed = 0;

  PRUint32 numChildren;
  threadHdr->GetNumChildren(&numChildren);
  NS_ASSERTION(numChildren, "Empty thread in view/db");
#ifndef NOT_POSTBOX
  // I'm not sure why Postbox needs this numChildren == 1 but we seem to be
  // expanding threads that only have one child in ExpandAll when changing the sort order causing the same hdr
  // to get listed twice.
  if (!numChildren || numChildren == 1)
    return NS_OK;
#endif

#if NOT_POSTBOX
  numChildren--; // account for the existing thread root
#endif
  if (!InsertEmptyRows(viewIndex, numChildren))
    return NS_ERROR_OUT_OF_MEMORY;

  PRBool threadedView = m_viewFlags & nsMsgViewFlagsType::kThreadedDisplay &&
    !(m_viewFlags & nsMsgViewFlagsType::kGroupBySort);
  nsMsgXFViewThread *viewThread;
  if (threadedView)
    viewThread = static_cast<nsMsgXFViewThread*>(threadHdr);

#if NOT_POSTBOX
  for (i = 1; i <= numChildren; i++)
#else
  for (i = 0; i <= numChildren; i++)
#endif
  {
    nsCOMPtr<nsIMsgDBHdr> msgHdr;
    threadHdr->GetChildHdrAt(i, getter_AddRefs(msgHdr));

    if (msgHdr)
    {
      nsMsgKey msgKey;
      PRUint32 msgFlags;
      msgHdr->GetMessageKey(&msgKey);
      msgHdr->GetFlags(&msgFlags);
#if NOT_POSTBOX
      PRUint8 level = (threadedView) ? viewThread->ChildLevelAt(i) : 1;
#else
      PRUint8 level = 1;
#endif
      SetMsgHdrAt(msgHdr, viewIndex, msgKey, msgFlags & ~MSG_VIEW_FLAGS, 
                  level);
      (*pNumListed)++;
      viewIndex++;
    }
  }
  SortKeys(m_sortType, m_sortOrder, startOfThreadViewIndex + 1, startOfThreadViewIndex + 1 + *pNumListed);
  return NS_OK;
}

// if nothing selected, return an NS_ERROR
NS_IMETHODIMP
nsMsgSearchDBView::GetHdrForFirstSelectedMessage(nsIMsgDBHdr **hdr)
{
  NS_ENSURE_ARG_POINTER(hdr);
  PRInt32 index;

  if (!mTreeSelection)
  {
    // We're in standalone mode, so use the message view index to get the header
    // We can't use the key here because we don't have an m_db
    index = m_currentlyDisplayedViewIndex;
  }
  else
  {
    nsresult rv = mTreeSelection->GetCurrentIndex(&index);
    NS_ENSURE_SUCCESS(rv, rv);
  }

  return GetMsgHdrForViewIndex(index, hdr);
}

NS_IMETHODIMP
nsMsgSearchDBView::OpenWithHdrs(nsISimpleEnumerator *aHeaders,
                                nsMsgViewSortTypeValue aSortType,
                                nsMsgViewSortOrderValue aSortOrder,
                                nsMsgViewFlagsTypeValue aViewFlags,
                                PRInt32 *aCount)
{
  if (aViewFlags & nsMsgViewFlagsType::kGroupBySort)
    return nsMsgGroupView::OpenWithHdrs(aHeaders, aSortType, aSortOrder, 
                                        aViewFlags, aCount);
#ifdef NO_POSTBOX
  m_sortType = aSortType;
  m_sortOrder = aSortOrder;
#else
  pbAdjustSortFieldsForPinning(aSortType, aSortOrder);
#endif
  m_viewFlags = aViewFlags;
  SaveSortInfo(m_sortType, m_sortOrder);

  PRBool hasMore;
  nsCOMPtr<nsISupports> supports;
  nsCOMPtr<nsIMsgDBHdr> msgHdr;
  nsCOMPtr<nsIMsgFolder> folder;
  nsresult rv = NS_OK;
  while (NS_SUCCEEDED(rv) && NS_SUCCEEDED(rv = aHeaders->HasMoreElements(&hasMore)) && hasMore)
  {
    rv = aHeaders->GetNext(getter_AddRefs(supports));
    if (NS_SUCCEEDED(rv) && supports)
    {
      msgHdr = do_QueryInterface(supports);
      msgHdr->GetFolder(getter_AddRefs(folder));
      AddHdrFromFolder(msgHdr, folder); 
    }
  }
  *aCount = m_keys.GetSize();
  return rv;
}

nsresult
nsMsgSearchDBView::GetFolderFromMsgURI(const char *aMsgURI, nsIMsgFolder **aFolder)
{
  nsCOMPtr <nsIMsgMessageService> msgMessageService;
  nsresult rv = GetMessageServiceFromURI(nsDependentCString(aMsgURI), getter_AddRefs(msgMessageService));
  NS_ENSURE_SUCCESS(rv,rv);
  
  nsCOMPtr <nsIMsgDBHdr> msgHdr;
  rv = msgMessageService->MessageURIToMsgHdr(aMsgURI, getter_AddRefs(msgHdr));
  NS_ENSURE_SUCCESS(rv,rv);
  
  return msgHdr->GetFolder(aFolder);
}

