/* This Source Code Form is subject to the terms of the Mozilla Public
 * License, v. 2.0. If a copy of the MPL was not distributed with this file,
 * You can obtain one at http://mozilla.org/MPL/2.0/. */

#import "mozNotificationCenterDelegate.h"
#import "ObserverPair.h"

#include "nsIObserver.h"
#include "nsIXULAppInfo.h"
#include "nsIStringBundle.h"
#include "nsIDOMWindow.h"
#include "nsObjCExceptions.h"
#include "nsIObserverService.h"

// Note that all calls to the Notification Center API are not made normally.
// We do this to maintain backwards compatibility.

id<FakeNSUserNotificationCenter> GetNotificationCenter() {
  Class c = NSClassFromString(@"NSUserNotificationCenter");
  return [c performSelector:@selector(defaultUserNotificationCenter)];
}

@implementation mozNotificationCenterDelegate

- (id)init
{
  NS_OBJC_BEGIN_TRY_ABORT_BLOCK_NIL;

  if ((self = [super init])) {
    mCenter = [GetNotificationCenter() retain];
    mKey = 0;
  }

  return self;

  NS_OBJC_END_TRY_ABORT_BLOCK_NIL;
}

- (void)dealloc
{
  NS_OBJC_BEGIN_TRY_ABORT_BLOCK;

  [mCenter release];

  [super dealloc];

  NS_OBJC_END_TRY_ABORT_BLOCK;
}

-(void) notifyWithTitle:(const nsAString&)aTitle
            description:(const nsAString&)aText
                    key:(PRUint32)aKey
                 cookie:(const nsAString&)aCookie;
{
  NS_OBJC_BEGIN_TRY_ABORT_BLOCK;

  Class unClass = NSClassFromString(@"NSUserNotification");
  id<FakeNSUserNotification> notification = [[unClass alloc] init];
  notification.title = [NSString stringWithCharacters:aTitle.BeginReading()
                                               length:aTitle.Length()];
  notification.informativeText = [NSString stringWithCharacters:aText.BeginReading()
                                                         length:aText.Length()];
  NSDictionary* clickContext = nil;
  clickContext = [NSDictionary dictionaryWithObjectsAndKeys:
                  [NSArray arrayWithObject:
                   [NSString stringWithCharacters:aCookie.BeginReading()
                                           length:aCookie.Length()]],
                  COOKIE_KEY,
                  nil];
  notification.userInfo = clickContext;
  notification.soundName = NSUserNotificationDefaultSoundName;
  notification.hasActionButton = NO;

  [GetNotificationCenter() deliverNotification:notification];

  NS_OBJC_END_TRY_ABORT_BLOCK;
}

- (PRUint32) addObserver:(nsIObserver *)aObserver
{
  NS_OBJC_BEGIN_TRY_ABORT_BLOCK_RETURN;
  // Postbox uses global notifications instead of window based notifications
  // in order to support activiating notifications.
  return 0;
  NS_OBJC_END_TRY_ABORT_BLOCK_RETURN(0);
}

- (void) forgetObserversForWindow:(nsIDOMWindow*)window
{
  NS_OBJC_BEGIN_TRY_ABORT_BLOCK;

  NS_OBJC_END_TRY_ABORT_BLOCK;
}

- (void) forgetObservers
{
  NS_OBJC_BEGIN_TRY_ABORT_BLOCK;

  NS_OBJC_END_TRY_ABORT_BLOCK;
}

// Implement NSUserNotificationCenterDelegate

- (void)userNotificationCenter:(id<FakeNSUserNotificationCenter>)center
        didDeliverNotification:(id<FakeNSUserNotification>)notification
{
  NS_OBJC_BEGIN_TRY_ABORT_BLOCK;

  // Internally we have two notifications, "alertclickcallback" and "alertfinished".
  // Notification Center does not tell us when the alert finished, but we always need
  // to send alert finished, or otherwise consider it finished in order to free up
  // observers. This is a conundrum. For now we'll do the unfortunately conservative
  // thing and not support actions so that we can consistently end alerts when they
  // are shown.

  NSDictionary* userInfo = notification.userInfo;

  NS_ASSERTION([userInfo valueForKey:COOKIE_KEY] != nil,
               "COOKIE_KEY not found!");

  // Notify any global listeners
  nsCOMPtr<nsIObserverService> os = do_GetService("@mozilla.org/observer-service;1");
  if (os) {
    NSString* cookie = [[userInfo valueForKey:COOKIE_KEY] objectAtIndex:0];

    nsAutoString tmp;
    tmp.SetLength([cookie length]);
    [cookie getCharacters:tmp.BeginWriting()];

    os->NotifyObservers(nsnull, "alertfinished", tmp.get());
  }

  NS_OBJC_END_TRY_ABORT_BLOCK;
}

- (void)userNotificationCenter:(id<FakeNSUserNotificationCenter>)center
       didActivateNotification:(id<FakeNSUserNotification>)notification
{
  NS_OBJC_BEGIN_TRY_ABORT_BLOCK;

  NSDictionary* userInfo = notification.userInfo;

  NS_ASSERTION([userInfo valueForKey:COOKIE_KEY] != nil,
               "COOKIE_KEY not found!");

  // Notify any global listeners
  nsCOMPtr<nsIObserverService> os = do_GetService("@mozilla.org/observer-service;1");
  if (os) {
    NSString* cookie = [[userInfo valueForKey:COOKIE_KEY] objectAtIndex:0];
    nsAutoString tmp;
    tmp.SetLength([cookie length]);
    [cookie getCharacters:tmp.BeginWriting()];
    os->NotifyObservers(nsnull, "alertclickcallback", tmp.get());
  }

  NS_OBJC_END_TRY_ABORT_BLOCK;
}

- (BOOL)userNotificationCenter:(id<FakeNSUserNotificationCenter>)center
     shouldPresentNotification:(id<FakeNSUserNotification>)notification
{
  // Respect OS choice.
  return NO;
}

@end
